# The main interfaces of the investor framework

__all__=[
    'DataCache',            'Fund',                 'Portfolio',    'Investor',
    'CurrencyConverter',    'CurrencyExchange',     'MarketIndex'
]

from .datacache              import DataCache
from .fund                   import Fund, KPI
from .portfolio              import Portfolio, PortfolioAggregator
from .monetary_time_series   import MarketIndex, CurrencyConverter, CurrencyExchange
from .investor               import Investor



import pathlib
import numpy
import pandas

from .. import Portfolio


class URIBalanceOrLedger(Portfolio):
    """
    Use this Portfolio class if you keep your balance and ledger in CSV files,
    local or remote -- on any URL supported by pandas.read_csv().

    The UI passes a sheet structure dict to this class constructor so it knows
    which tabs and columns contains the information it needs. The UI get this
    information from investor_ui_config.yaml file, under
    portfolio.params.sheetStructure, which in turns define where in the
    spreadsheet are balance and ledger information. Here is an example
    for a investor_ui_config.yaml:

    portfolio:
        - type: !!python/name:investor.portfolios.uri.URIBalanceOrLedger ''
          params:
            # The URI can be a local file or even a remote http(s):// URL.
            # Anything supported by Pandas.read_csv()
            URI: balances.txt
            kind: traderbot_balance
            sheetStructure:
                separator: "|"
                # In here you describe how the BALANCE and LEDGER data is
                # organized in sheets and columns
                balance:
                    columns:
                        # Time is in column called 'time'
                        time: time

                        # Name of funds on each row is under this column
                        fund: fund

                        # Column called 'Saldo USD' contains values in 'USD' and so on.
                        monetary:
                            - currency:     USD
                              name:         Saldo USD
    """

    def __init__(self, URI, kind, sheetStructure, base='.', cache=None, refresh=False):

        if '://' in URI:
            self.URI=URI
        else:
            self.URI=pathlib.Path(URI).expanduser()
            if not self.URI.is_absolute():
                self.URI=(
                    (pathlib.Path(base).parent.resolve() / self.URI)
                    .resolve()
                    .as_uri()
                )

        self.kind = kind
        self.sheetStructure = sheetStructure

        self._balance = None
        self._ledger = None

        super().__init__(
            kind       = f'uri•{kind}',
            id         = str(self.URI),   # extract just full name from Path object
            cache      = cache,
            refresh    = refresh
        )



    ############################################################################
    ##
    ## Portfolio interface methods.
    ##
    ## Concrete implementation of abstract virtual methods from Portfolio class
    ##
    ############################################################################


    @property
    def has_balance(self):
        return 'balance' in self.sheetStructure



    @property
    def has_ledger(self):
        return 'ledger' in self.sheetStructure



    def refreshData(self):
        """
        Force data refresh from source URL and return a DataFrame which columns
        where already renamed and filtered, ready to be cached. But data still
        needs to be processed by self.processData().

        The internal variables _balance or _ledger will be updated.
        """
        if self.has_balance:
            prop='balance'
        elif self.has_ledger:
            prop='ledger'
        else:
            raise NameError('Either balance or ledger sheet structure must be defined.')

        columnsProfile=self.sheetStructure[prop]['columns']

        # Normalize all column names
        renamer={m['name']: m['currency'] for m in columnsProfile['monetary']}
        renamer.update(
            {
                columnsProfile[k]: k
                for k in columnsProfile.keys() if k!='monetary'
            }
        )

        df=(
            pandas.read_csv(
                filepath_or_buffer = self.URI,
                sep = (
                    self.sheetStructure['separator']
                    if 'separator' in self.sheetStructure
                    else ','
                )
            )
            .rename(columns=renamer)
            .pipe(
                lambda table: table.drop(
                    # Drop all columns that doesn't matter
                    columns=list(
                        set(table.columns) -
                        set(list(renamer.values()))
                    )
                )
            )
        )

        setattr(self,f'_{prop}',df)

        return getattr(self,f'_{prop}')



    def processData(self):
        if self.has_balance:
            self.processSheetData('balance')
        if self.has_ledger:
            self.processSheetData('ledger')



    ############################################################################
    ##
    ## Internal methods
    ##
    ############################################################################

    def processSheetData(self, prop):
        def ddebug(table):
            self.logger.debug(table)
            return table

        sheet=getattr(self,f'_{prop}')
        columnsProfile=self.sheetStructure[prop]['columns']
        monetaryColumns=[c['currency'] for c in columnsProfile['monetary']]

        # Timestamps whose time part is 0 will be shifted with naiveTimeShift
        naiveTimeShift=12*3600
        if prop=='balance':
            naiveTimeShift+=3*60

        sheet=(
            sheet

            # Normalize NaNs
            .fillna(pandas.NA)
            .replace('#N/A',pandas.NA)

            # Remove rows that don't have fund names or monetary values
            .dropna(subset=monetaryColumns, how='all')
            .dropna(subset=['fund'])

            # Optimize and be gentle with storage
            .astype(
                dict(
                    fund = 'category'
                )
            )

            .assign(
                # Convert Date/Time to proper type
                time=lambda table: Portfolio.normalizeTime(
                    pandas.to_datetime(
                        table.time,
                        format='mixed',
                        yearfirst=True,
                        utc=False
                    ),
                    naiveTimeShift
                )
            )
        )

        # Handle monetary columns, remove currency symbols and make them numbers
        remove=['R$ ','$','€',',']
        for c in columnsProfile['monetary']:
            if sheet[c['currency']].dtype==numpy.dtype('O'):
                ## Remove currency symbols and junk
                for r in remove:
                    sheet[c['currency']]=sheet[c['currency']].str.replace(r, '', regex=False)

                ## Make NaNs of empty ('') cells
                sheet[c['currency']]=sheet[c['currency']].str.replace(r'^\s*$','', regex=True)

                ## Convert to number
                sheet[c['currency']]=pandas.to_numeric(sheet[c['currency']]) #.astype(float)

        setattr(self,f'_{prop}',sheet)



    def __repr__(self):
        return '{klass}<{kind}>({balance_or_ledger}={URI})'.format(
            URI               = self.URI,
            balance_or_ledger = 'balance' if self.has_balance else 'ledger',
            klass             = type(self).__name__,
            kind              = self.kind
        )



    def to_markdown(self, title_prefix=None):
        """
        Provides a markdown representation of this object to be used in the
        "Portfolio Components and Information" tab of Investorzilla's UI.
        """
        title="{klass}({kind})".format(klass=type(self).__name__,kind=self.kind)

        if '://' in str(self.URI):
            URI = f"[remote URL]({self.URI})"
        else:
            URI = f"`{self.URI}`"

        nonMonetary=set(['time','fund','comment'])

        if self.has_balance:
            data=self.balance
        else:
            data=self.ledger

        body=[
            "- {balance_or_ledger} from {URI}".format(
                balance_or_ledger = 'balance' if self.has_balance else 'ledger',
                URI=URI
            ),
            "- assets: `{}`".format('` · `'.join(data.fund.unique())),
            "- currencies: `{}`".format('` · `'.join(set(data.columns)-nonMonetary)),
            f"- from `{data.time.min()}` to `{data.time.max()}`"
        ]

        body='\n'.join(body)

        if title_prefix is None:
            return (title,body)
        else:
            return f"{title_prefix} {title}\n{body}"






import os
import subprocess
import sys
import time
import importlib
import webbrowser
import investorzilla

def main():
    # Getting path to python executable (full path of deployed python on Windows)
    executable = sys.executable

    # Running streamlit server in a subprocess and writing to log file
    proc = subprocess.Popen(
        [
            executable,
            "-m",
            "streamlit",
            "run",
            str(importlib.resources.files(investorzilla) / 'streamlit_ui.py'),
            # The following option appears to be necessary to correctly start the streamlit server,
            # but it should start without it. More investigations should be carried out.
            "--server.headless=true",
            "--global.developmentMode=false",
        ],
        # stdin=subprocess.PIPE,
        # stdout=subprocess.PIPE,
        # stderr=subprocess.STDOUT,
        # text=True,
    )

    # proc.stdin.close()

    # Force the opening (does not open automatically) of the browser tab after a brief delay to let
    # the streamlit server start.
    time.sleep(2)
    webbrowser.open("http://localhost:8501")

    # while True:
    #     s = proc.stdout.read()
    #     if not s:
    #         break
    #     print(s, end="")

    proc.wait()


if "__main__" in __name__:
    main()



import datetime
import logging

# Dependencies available via OS packages:
# pip3 install pandas pyyaml sqlalchemy pandas_datareader

# Other dependencies:
# pip3 install streamlit google-api-python-client

import streamlit
import investorzilla


class StreamlitInvestorzillaApp:

    defaultRefreshMap=dict(
        zip(
            investorzilla.Investor.domains,
            len(investorzilla.Investor.domains)*[False]
        )
    )

    def prepare_logging(self,level=logging.INFO):
        # Switch between INFO/DEBUG while running in production/developping:

        # Configure logging for Investor

        loggers=[
            logging.getLogger('__main__'),
            logging.getLogger('investorzilla'),
            # logging.getLogger('sqlite')
        ]

        if loggers[0].handlers:  # logger is already setup, don't setup again
            return loggers[0]

        FORMATTER = logging.Formatter("%(asctime)s|%(levelname)s|%(name)s|%(message)s")
        HANDLER = logging.StreamHandler()
        HANDLER.setFormatter(FORMATTER)

        for logger in loggers:
            logger.addHandler(HANDLER)
            logger.setLevel(level)

        streamlit.session_state.logger=loggers[0]
        return streamlit.session_state.logger



    def __init__(self, refresh=False):
        self.prepare_logging(level=logging.INFO)

        streamlit.set_page_config(
            layout="wide",
            page_title='Investorzilla',
            menu_items={
                "Report a Bug": 'https://github.com/avibrazil/investorzilla/issues',
                "About": '\n'.join([
                    '# [Investorzilla](https://github.com/avibrazil/investorzilla)',
                    'Brought to you by **[Avi Alkalay](https://Avi.Alkalay.NET/)**'
                ])
            }
        )
        with streamlit.sidebar:
            # Get the kind of refresh user wants, if any
            self.refreshMap=self.interact_refresh()

        # investor    refresh
        #   None         X.     => load
        #     X          True.  => load
        #     X.         False. => reuse
        #     None.      False. => load


        if ('investor' not in streamlit.session_state) or (True in self.refreshMap.values()):
            streamlit.session_state.investor=investorzilla.Investor(
                'investorzilla.yaml',
                self.refreshMap
            )

        # Make a preliminary internal fund from the portfolio for display purposes
        streamlit.session_state.investor.portfolio.makeInternalFund(
            streamlit.session_state.investor.exchange
        )

        with streamlit.sidebar:
            # Put controls in the sidebar
            self.interact_assets()
            self.interact_exclude_assets()
            self.interact_start_end()
            self.interact_currencies()
            self.interact_benchmarks()
            self.interact_periods()

        # Render main content with plots and tables
        self.update_content()



    def prepare_fund(self):
        """
        Generate a virtual fund (shares and share value) based on portfolio
        items (assets) selected in sidebar.
        """

        fundset=None
        if 'interact_assets' in streamlit.session_state:
            assets=streamlit.session_state.interact_assets

            if 'ALL' in assets:
                assets.remove('ALL')

        if assets is None or len(assets)==0:
            assets=streamlit.session_state.investor.portfolio.assets()
            assets=[f[0] for f in assets]

        if 'interact_no_assets' in streamlit.session_state:
            assets=list(
                set(assets) -
                set(streamlit.session_state.interact_no_assets)
            )

        # Make a virtual fund (shares and share value) from selected assets
        streamlit.session_state.fund=streamlit.session_state.investor.portfolio.getFund(
            subset           = assets,
            currencyExchange = streamlit.session_state.investor.exchange
        )

        streamlit.session_state.fund.setName(top=4)

        p=investorzilla.Fund.periodPairs[streamlit.session_state.interact_periods]

        self.start = (
            streamlit.session_state.interact_start_end[0]
            if len(streamlit.session_state.interact_start_end)>0
            else streamlit.session_state.investor.portfolio.fund.start
        )
        self.end   = (
            streamlit.session_state.interact_start_end[1]
            if len(streamlit.session_state.interact_start_end)>1
            else streamlit.session_state.investor.portfolio.fund.end
        )

        # Now that we have a fund, create periodic reports
        self.reportRagged=streamlit.session_state.fund.periodicReport(
            benchmark  = streamlit.session_state.interact_benchmarks['obj'],
            start      = self.start,
            end        = self.end,
        )

        self.reportPeriodic=streamlit.session_state.fund.periodicReport(
            period     = p['period'],
            benchmark  = streamlit.session_state.interact_benchmarks['obj'],
            start      = self.start,
            end        = self.end,
        )

        self.reportMacroPeriodic=streamlit.session_state.fund.periodicReport(
            period     = p['macroPeriod'],
            benchmark  = streamlit.session_state.interact_benchmarks['obj'],
            start      = self.start,
            end        = self.end,
        )

        self.report=streamlit.session_state.fund.report(
            precomputedPeriodicReport      = self.reportPeriodic,
            precomputedMacroPeriodicReport = self.reportMacroPeriodic,
            period     = streamlit.session_state.interact_periods,
            benchmark  = streamlit.session_state.interact_benchmarks['obj'],
            start      = self.start,
            end        = self.end,
        )



    def update_content(self):
        """
        Render the report
        """

        streamlit.session_state.investor.currency=streamlit.session_state.interact_currencies

        self.prepare_fund()

        (
            tab_performance,
            tab_wealth,
            tab_contributions,
            tab_summary,
            tab_shares,
            tab_currencies,
            tab_portfolio
        ) = streamlit.tabs(
            [
                "📈 Performance",
                "📈 Wealth",
                "📶 Per Asset Contributions",
                "💼 Portfolio Summary",
                "📶 Fund Shares inspector",
                "🔁 Currencies inspector",
                "💼 Portfolio Components and Information",
            ]
        )

        with tab_performance:
            self.render_performance_page()

        with tab_wealth:
            self.render_wealth_page()

        with tab_contributions:
            self.render_contributions_page()

        with tab_summary:
            self.render_summary_page()

        with tab_shares:
            self.render_shares_page()

        with tab_currencies:
            self.render_currencies_page()

        with tab_portfolio:
            self.render_portfolio_page()

        streamlit.caption('Report by [investorzilla](https://github.com/avibrazil/investorzilla).')



    def render_summary_page(self):
        # Render title
        streamlit.title(streamlit.session_state.fund.name)

        streamlit.dataframe(streamlit.session_state.fund.describe())



    def render_contributions_page(self):
        # Render title
        streamlit.title(streamlit.session_state.fund.name)

        kpis=[
            investorzilla.KPI.PERIOD_GAIN,
            investorzilla.KPI.GAINS,
            investorzilla.KPI.BALANCE,
            investorzilla.KPI.SAVINGS,
            investorzilla.KPI.MOVEMENTS
        ]

        col1, col2 = streamlit.columns(2)

        col1.radio(
            "Select KPI to show the contribution of each Asset",
            kpis,
            key='kpi_contributions',
            horizontal=True,
        )

        col2.selectbox(
           "Select point in time",
            self.reportPeriodic.index.sort_values(ascending=False),
            key="pointintime_contributions"
        )

        streamlit.altair_chart(
            use_container_width=True,
            altair_chart=streamlit.session_state.fund.assetContributionPlot(
                pointInTime=streamlit.session_state.pointintime_contributions,
                kpi=streamlit.session_state.kpi_contributions,
                period=investorzilla.Fund.periodPairs[streamlit.session_state.interact_periods]['period'],
            ).interactive()
        )



    def render_wealth_page(self):
        p=investorzilla.Fund.periodPairs[streamlit.session_state.interact_periods]

        # Render title
        streamlit.title(streamlit.session_state.fund.name)

        # Render main metrics
        # streamlit.header('Main Metrics')

        if streamlit.session_state.interact_benchmarks['obj'].currency!=streamlit.session_state.fund.currency:
            streamlit.warning('Fund and Benchmark have different currencies. Benchmark comparisons won’t make sense.')

        col1, col2, col3 = streamlit.columns(3)

        label='{kpi}: current {p[periodLabel]} and {p[macroPeriodLabel]}'

        # Balance
        col1.metric(
            label='current {} & {}'.format(investorzilla.KPI.BALANCE,investorzilla.KPI.SAVINGS),
            value='${:0,.2f}'.format(self.reportPeriodic.iloc[-1][investorzilla.KPI.BALANCE]),
            delta='{sign}${value:0,.2f}'.format(
                value=abs(self.reportMacroPeriodic.iloc[-1][investorzilla.KPI.SAVINGS]),
                sign=('-' if self.reportMacroPeriodic.iloc[-1][investorzilla.KPI.SAVINGS]<0 else '')
            ),
            help='Current balance compared to all your savings'
        )

        col2.metric(
            label=investorzilla.KPI.BALANCE_OVER_SAVINGS,
            value='{:6.2f}%'.format(100*self.reportPeriodic[investorzilla.KPI.BALANCE_OVER_SAVINGS].iloc[-1]),
            help='How many times your balance is bigger than your savings'
        )

        # Latest average movements (power of saving)
        col3.metric(
            label=f"{investorzilla.KPI.MOVEMENTS}: last {p['macroPeriodLabel']} median and mean for a {p['periodLabel']}",
            value='${:0,.2f}'.format(
                self.reportPeriodic[investorzilla.KPI.MOVEMENTS]
                .tail(investorzilla.Fund.div_offsets(p['macroPeriod'],p['period']))
                .median()
            ),
            delta='${:0,.2f}'.format(
                self.reportPeriodic[investorzilla.KPI.MOVEMENTS]
                .tail(investorzilla.Fund.div_offsets(p['macroPeriod'],p['period']))
                .mean()
            ),
            help='Long term tendency of saving money'
        )

        col1, col2 = streamlit.columns(2)

        with col1:
            streamlit.line_chart(
                use_container_width=True,
                data=streamlit.session_state.fund.wealthPlot(
                    benchmark=streamlit.session_state.interact_benchmarks['obj'],
                    precomputedReport=self.reportRagged,
                    type='raw'
                )
            )

            streamlit.line_chart(
                use_container_width=True,
                data=self.reportRagged,
                y=investorzilla.KPI.BALANCE_OVER_SAVINGS
            )

        with col2:
            streamlit.altair_chart(
                use_container_width=True,
                altair_chart=streamlit.session_state.fund.genericPeriodicPlot(
                    kpi=investorzilla.KPI.MOVEMENTS,
                    periodPair=streamlit.session_state.interact_periods,
                    type='altair',
                    precomputedReport=self.reportPeriodic
                ).interactive()
            )

        table_styles=[
            dict(selector="td", props="font-size: 0.8em; text-align: right"),
            dict(selector="th", props="font-size: 0.8em; "),
            dict(selector='tr:hover', props='background-color: yellow')
        ]

        streamlit.header('Wealth Evolution')

        wealth_benchmarks=[
            investorzilla.KPI.BALANCE,
            investorzilla.KPI.BALANCE_OVER_SAVINGS,
            investorzilla.KPI.GAINS,
            investorzilla.KPI.SAVINGS,
            investorzilla.KPI.MOVEMENTS
        ]

        streamlit.multiselect(
            '',
            options=wealth_benchmarks,
            default=wealth_benchmarks,
            key='kpi_wealth'
        )

        streamlit.markdown(
            investorzilla.Fund.format(
                investorzilla.Fund.filter(
                    self.report,
                    kpi=streamlit.session_state.kpi_wealth
                )
            )
            .set_table_styles(table_styles)

            .to_html(),
            unsafe_allow_html=True
        )

        # Render footer stats
        streamlit.markdown(
            'Most recent porfolio data is **{}**'.format(
                streamlit.session_state.investor.portfolio.asof
            )
        )

        streamlit.markdown(
            'Graph data between **{}** and **{}**'.format(self.start,self.end)
        )



    def render_performance_page(self):
        p=investorzilla.Fund.periodPairs[streamlit.session_state.interact_periods]

        # Render title
        streamlit.title(streamlit.session_state.fund.name)

        # Render period slider
        # self.interact_start_end()

        # Render main metrics
        # streamlit.header('Main Metrics')

        if streamlit.session_state.interact_benchmarks['obj'].currency!=streamlit.session_state.fund.currency:
            streamlit.warning('Fund and Benchmark have different currencies. Benchmark comparisons won’t make sense.')

        col1, col2, col3, col4 = streamlit.columns(4)

        label='{kpi}: current {p[periodLabel]} and {p[macroPeriodLabel]}'

        # Rate of return
        col1.metric(
            label=label.format(p=p,kpi=investorzilla.KPI.RATE_RETURN),
            value='{:6.2f}%'.format(100*self.reportPeriodic.iloc[-1][investorzilla.KPI.RATE_RETURN]),
            delta='{:6.2f}%'.format(100*self.reportMacroPeriodic.iloc[-1][investorzilla.KPI.RATE_RETURN]),
        )

        # Latest average rate of return
        col2.metric(
            label=f"{investorzilla.KPI.RATE_RETURN}: {p['periodLabel']} median and mean over last {p['macroPeriodLabel']}",
            value='{:6.2f}%'.format(100 * (
                    self.reportPeriodic[investorzilla.KPI.RATE_RETURN]
                    .tail(investorzilla.Fund.div_offsets(p['macroPeriod'],p['period']))
                    .median()
                )
            ),
            delta='{:6.2f}%'.format(100 * (
                    self.reportPeriodic[investorzilla.KPI.RATE_RETURN]
                    .tail(investorzilla.Fund.div_offsets(p['macroPeriod'],p['period']))
                    .mean()
                )
            ),
        )

        # Gain
        col3.metric(
            label=label.format(p=p,kpi=investorzilla.KPI.PERIOD_GAIN),
            value='${:0,.2f}'.format(self.reportPeriodic.iloc[-1][investorzilla.KPI.PERIOD_GAIN]),
            delta='{sign}${value:0,.2f}'.format(
                value=abs(self.reportMacroPeriodic.iloc[-1][investorzilla.KPI.PERIOD_GAIN]),
                sign=('-' if self.reportMacroPeriodic.iloc[-1][investorzilla.KPI.PERIOD_GAIN]<0 else '')
            )
        )

        # Average Gain
        col4.metric(
            label=f"{investorzilla.KPI.PERIOD_GAIN}: {p['periodLabel']} median and mean over last {p['macroPeriodLabel']}",
            value='${:0,.2f}'.format(
                self.reportPeriodic[investorzilla.KPI.PERIOD_GAIN]
                .tail(investorzilla.Fund.div_offsets(p['macroPeriod'],p['period']))
                .median()
            ),
            delta='${:0,.2f}'.format(
                self.reportPeriodic[investorzilla.KPI.PERIOD_GAIN]
                .tail(investorzilla.Fund.div_offsets(p['macroPeriod'],p['period']))
                .mean()
            ),
        )


        col1, col2 = streamlit.columns(2)

        with col1:
            streamlit.header('Performance', divider='red')
            streamlit.line_chart(
                streamlit.session_state.fund.performancePlot(
                    benchmark=streamlit.session_state.interact_benchmarks['obj'],
                    precomputedReport=self.reportRagged,
                    type='raw'
                )
            )

            streamlit.altair_chart(
                use_container_width=True,
                altair_chart=streamlit.session_state.fund.rateOfReturnPlot(
                    precomputedReport=self.reportPeriodic,
                    type='altair'
                ).interactive()
            )

        with col2:
            streamlit.header('Gains', divider='red')
            streamlit.altair_chart(
                use_container_width=True,
                altair_chart=streamlit.session_state.fund.genericPeriodicPlot(
                    kpi=investorzilla.KPI.PERIOD_GAIN,
                    periodPair=streamlit.session_state.interact_periods,
                    type='altair',
                    precomputedReport=self.reportPeriodic
                ).interactive()
            )

            streamlit.altair_chart(
                use_container_width=True,
                altair_chart=streamlit.session_state.fund.genericPeriodicPlot(
                    kpi=investorzilla.KPI.RATE_RETURN,
                    periodPair=streamlit.session_state.interact_periods,
                    type='altair',
                    precomputedReport=self.reportPeriodic
                ).interactive()
            )

        table_styles=[
            dict(selector="td", props="font-size: 0.8em; text-align: right"),
            dict(selector="th", props="font-size: 0.8em; "),
            dict(selector='tr:hover', props='background-color: yellow')
        ]

        streamlit.header('Performance')
        streamlit.markdown(
            "Benchmark is **{}**."
            .format(streamlit.session_state.interact_benchmarks['obj'])
        )

        performance_benchmarks=[
            investorzilla.KPI.RATE_RETURN,
            investorzilla.KPI.BENCHMARK_RATE_RETURN,
            investorzilla.KPI.BENCHMARK_EXCESS_RETURN,
            investorzilla.KPI.PERIOD_GAIN
        ]

        streamlit.multiselect(
            '',
            options=performance_benchmarks,
            default=performance_benchmarks,
            key='kpi_performance'
        )

        streamlit.markdown(
            investorzilla.Fund.format(
                investorzilla.Fund.filter(
                    self.report,
                    kpi=streamlit.session_state.kpi_performance
                )
            )
            .set_table_styles(table_styles)

            .to_html(),
            unsafe_allow_html=True
        )

        # Render footer stats
        streamlit.markdown(
            'Most recent porfolio data is **{}**'.format(
                streamlit.session_state.investor.portfolio.asof
            )
        )

        streamlit.markdown(
            'Graph data between **{}** and **{}**'.format(self.start,self.end)
        )



    def render_currencies_page(self):
        streamlit.title(f"1 {streamlit.session_state.investor.exchange.currency} in other currencies")
        streamlit.dataframe(
            (
                (1/streamlit.session_state.investor.exchange.data)
                # Deliver time information in user’s timezone
                .assign(
                    time=lambda table: (
                        table.index
                        .tz_convert(
                            datetime.datetime.now(datetime.timezone.utc)
                            .astimezone()
                            .tzinfo
                        )
                    ),
                )
                .set_index('time')
                .sort_index(ascending=False)
            ),
            use_container_width=True,
            column_config={
                c: streamlit.column_config.NumberColumn(format="%.16f")
                for c in streamlit.session_state.investor.exchange.data.columns
            }
        )



    def render_shares_page(self):
        streamlit.title(streamlit.session_state.fund.name)
        streamlit.dataframe(
            (
                streamlit.session_state.fund.shares
                .assign(
                    time=lambda table: table.index.tz_convert(datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo),
                    balance=lambda table: (
                        table[investorzilla.KPI.SHARES] *
                        table[investorzilla.KPI.SHARE_VALUE]
                    )
                )
                .set_index('time')
            ),
            use_container_width=True,
            column_config={
                investorzilla.KPI.SHARES: streamlit.column_config.NumberColumn(
                    help="Number of shares of the virtual fund formed by selected assets, in that point of time",
                ),
                investorzilla.KPI.SHARE_VALUE: streamlit.column_config.NumberColumn(
                    help=f"The value of each share, in {streamlit.session_state.fund.currency}",
                    # format=investorzilla.Fund.formatters[investorzilla.KPI.SHARE_VALUE]['format'],
                    format="$%.6f"
                ),
                investorzilla.KPI.BALANCE: streamlit.column_config.NumberColumn(
                    help=f"Balance in that point of time, in {streamlit.session_state.fund.currency}",
                    format="$%.2f"
                )
            }
        )



    def render_portfolio_page(self):
        with streamlit.expander('Personal Portfolio'):
            streamlit.markdown(streamlit.session_state.investor.portfolio.to_markdown(title_prefix='##'))

        with streamlit.expander('Currency Converters'):
            for c in streamlit.session_state.investor.currency_converters:
                streamlit.markdown(c['obj'].to_markdown(title_prefix='###'))

        with streamlit.expander('Market Indexes'):
            for b in streamlit.session_state.investor.benchmarks:
                streamlit.markdown(b['obj'].to_markdown(title_prefix='###'))



    # All the interact_* methods manage widgets in the Streamlit UI

    def interact_assets(self):
        streamlit.multiselect(
            label     = 'Select assets to make a fund',
            options   = (
                ['ALL']+
                [
                    x[0]
                    for x in streamlit.session_state.investor.portfolio.assets()
                ]
            ),
            help      = 'Shares and share value will be computed for the union of selected assets',
            key       = 'interact_assets'
        )



    def interact_exclude_assets(self):
        streamlit.multiselect(
            label     = 'Except assets',
            options   = [
                x[0]
                for x in streamlit.session_state.investor.portfolio.assets()
            ],
            help      = 'Exclude assets selected here',
            key       = 'interact_no_assets'
        )



    def interact_currencies(self):
        currencies=streamlit.session_state.investor.exchange.currencies()

        # Find the index of default currency
        for i in range(len(currencies)):
            if currencies[i]==streamlit.session_state.investor.config['currency']:
                break

        streamlit.radio(
            label     = 'Convert all to currency',
            options   = currencies,
            index     = i,
            help      = 'Everything will be converted to selected currency',
            key       = 'interact_currencies'
        )



    def interact_benchmarks(self):
        streamlit.radio(
            label       = 'Select a benchmark to compare with',
            options     = streamlit.session_state.investor.benchmarks,
            format_func = lambda bench: str(bench['obj']),
            help        = 'Funds will be compared to the selected benchmark',
            key         = 'interact_benchmarks'
        )



    def interact_start_end(self):
        inv=streamlit.session_state.investor

        key='relevant period'
        defaults = dict(
            start = (
                inv.config[key]['start']
                if key in inv.config and 'start' in inv.config[key]
                else inv.portfolio.fund.start.to_pydatetime()
            ),
            end = (
                inv.config[key]['end']
                if key in inv.config and 'end' in inv.config[key]
                else inv.portfolio.fund.end.to_pydatetime()
            )
        )

        streamlit.date_input(
            label       = 'Report Period Range',
            help        = 'Report starting on date',
            min_value   = inv.portfolio.fund.start.to_pydatetime(),
            max_value   = inv.portfolio.fund.end.to_pydatetime(),
            value       = (defaults['start'],defaults['end']),
            format      = 'YYYY-MM-DD',
            key         = 'interact_start_end'
        )



    def interact_periods(self):
        streamlit.radio(
            label       = 'How to divide time',
            options     = investorzilla.Fund.getPeriodPairs(),
            # format_func = investorzilla.Fund.getPeriodPairLabel,
            index       = investorzilla.Fund.getPeriodPairs().index('month & year'), # the starting default
            help        = 'Pairs of period (as month) and macro period (as year)',
            key         = 'interact_periods'
        )



    def interact_refresh(self):
        self.refreshMap=self.defaultRefreshMap.copy()

        streamlit.subheader('Refresh data')

        col1, col2, col3 = streamlit.columns(3)

        if col1.button(
            label       = 'Portfolio',
            help        = 'Invalidate cache and update your Portfolio data from the Internet'
        ):
            self.refreshMap['portfolio']=True

        if col2.button(
            label       = 'Market',
            help        = 'Invalidate cache and update Market Indexes and Currency Converters data from the Internet'
        ):
            self.refreshMap['currency_converters']=True
            self.refreshMap['benchmarks']=True

        if col3.button(
            label       = 'Both',
            help        = 'Invalidate cache and update all data from the Internet'
        ):
            self.refreshMap['currency_converters']=True
            self.refreshMap['benchmarks']=True
            self.refreshMap['portfolio']=True

        return self.refreshMap



StreamlitInvestorzillaApp(refresh=False)



import pandas
import pandas_datareader

from .. import MarketIndex


class FREDMarketIndex(MarketIndex):
    """
    Any market index from https://fred.stlouisfed.org/categories/32255

    Use the tag in the URL in the `name` parameter. For example, in the page above,
    the `S&P 500` has URL https://fred.stlouisfed.org/series/SP500, so use name=SP500.

    """
    def __init__(self, name, isRate=False, cache=None, refresh=False):
        super().__init__(kind='DataReaderFRED', id=name, currency='USD', isRate=isRate, cache=cache, refresh=refresh)


    def refreshData(self):
        self.data = pandas_datareader.data.DataReader([self.id], 'fred','1900-01-01')

        # Put a standard column name
        self.data.rename(columns={self.id: 'value'}, inplace=True)

#         # Rename index to our standard
#         self.data.index.name='time'

        # Make the index a regular column for caching purposes
        self.data.reset_index(inplace=True)



    def processData(self):
        # Convert time to a new column
        self.data['time']=pandas.to_datetime(self.data.DATE)

        # Set it as the index
        self.data.set_index('time', inplace=True)
        self.data.sort_index(inplace=True)

        # Drop old column
        self.data.drop('DATE', axis=1)

        self.data.fillna(method='ffill', axis=0, inplace=True)

        # Compute rate from daily values
        self.data['rate']=self.data['value']/self.data.shift()['value']-1


